from .base import CoreMetadata, RoleTypes, RuleModel


class DomainMetadata(CoreMetadata):
    role: RoleTypes = RoleTypes.domain_expert
    creator: str | list[str]


class DomainRules(RuleModel):
    metadata: DomainMetadata



"""This module contains the definition of `TransformationRules` pydantic model and all
its sub-models and validators.
"""

from __future__ import annotations

import math
import sys
from datetime import datetime
from functools import wraps
from typing import ClassVar, TypeAlias

import pandas as pd
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    HttpUrl,
    constr,
)
from pydantic.fields import FieldInfo
from rdflib import Namespace

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

#     "Class",
#     "Classes",
#     "Instance",
#     "CoreMetadata",
#     "Prefixes",
#     "Property",
#     "Properties",
#     "Resource",
#     "Rules",

METADATA_VALUE_MAX_LENGTH = 5120


def replace_nan_floats_with_default(values: dict, model_fields: dict[str, FieldInfo]) -> dict:
    output = {}
    for field_name, value in values.items():
        is_nan_float = isinstance(value, float) and math.isnan(value)
        if not is_nan_float:
            output[field_name] = value
            continue
        if field_name in model_fields:
            output[field_name] = model_fields[field_name].default
        else:
            # field_name may be an alias
            source_name = next((name for name, field in model_fields.items() if field.alias == field_name), None)
            if source_name:
                output[field_name] = model_fields[source_name].default
            else:
                # Just pass it through if it is not an alias.
                output[field_name] = value
    return output


def skip_field_validator(validators_field):
    def decorator(func):
        @wraps(func)
        def wrapper(cls, value, values):
            if isinstance(values, dict):
                to_skip = values.get(validators_field, set())
            else:
                try:
                    to_skip = values.data.get(validators_field, set())
                except Exception:
                    to_skip = set()

            if "all" in to_skip or func.__name__ in to_skip:
                return value
            return func(cls, value, values)

        return wrapper

    return decorator


def skip_model_validator(validators_field):
    def decorator(func):
        @wraps(func)
        def wrapper(self):
            to_skip = getattr(self, validators_field, set())
            if "all" in to_skip or func.__name__ in to_skip:
                return self

            return func(self)

        return wrapper

    return decorator


def _get_required_fields(model: type[BaseModel], use_alias: bool = False) -> set[str]:
    """Get required fields from a pydantic model.

    Parameters
    ----------
    model : type[BaseModel]
        Pydantic data model
    use_alias : bool, optional
        Whether to return field alias name, by default False

    Returns
    -------
    list[str]
        List of required fields
    """
    required_fields = set()
    for name, field in model.model_fields.items():
        if not field.is_required():
            continue

        alias = getattr(field, "alias", None)
        if use_alias and alias:
            required_fields.add(alias)
        else:
            required_fields.add(name)
    return required_fields


########################################################################################
### These highly depend on CDF API endpoint limitations we need to keep them updated ###
########################################################################################
more_than_one_none_alphanumerics_regex = r"([_-]{2,})"

prefix_compliance_regex = r"^([a-zA-Z]+)([a-zA-Z0-9]*[_-]{0,1}[a-zA-Z0-9_-]*)([a-zA-Z0-9]*)$"
data_model_id_compliance_regex = r"^[a-zA-Z]([a-zA-Z0-9_]{0,253}[a-zA-Z0-9])?$"
cdf_space_compliance_regex = (
    r"(?!^(space|cdf|dms|pg3|shared|system|node|edge)$)(^[a-zA-Z][a-zA-Z0-9_-]{0,41}[a-zA-Z0-9]?$)"
)

view_id_compliance_regex = (
    r"(?!^(Query|Mutation|Subscription|String|Int32|Int64|Int|Float32|Float64|Float|"
    r"Timestamp|JSONObject|Date|Numeric|Boolean|PageInfo|File|Sequence|TimeSeries)$)"
    r"(^[a-zA-Z][a-zA-Z0-9_]{0,253}[a-zA-Z0-9]?$)"
)
dms_property_id_compliance_regex = (
    r"(?!^(space|externalId|createdTime|lastUpdatedTime|deletedTime|edge_id|"
    r"node_id|project_id|property_group|seq|tg_table_name|extensions)$)"
    r"(^[a-zA-Z][a-zA-Z0-9_]{0,253}[a-zA-Z0-9]?$)"
)


class_id_compliance_regex = r"(?!^(Class|class)$)(^[a-zA-Z][a-zA-Z0-9._-]{0,253}[a-zA-Z0-9]?$)"
property_id_compliance_regex = r"^(\*)|(?!^(Property|property)$)(^[a-zA-Z][a-zA-Z0-9._-]{0,253}[a-zA-Z0-9]?$)"

version_compliance_regex = r"^[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])?$"
########################################################################################
########################################################################################

Prefix: TypeAlias = str
ExternalId: TypeAlias = str
Space: TypeAlias = str
Description: TypeAlias = constr(min_length=1, max_length=1024)  # type: ignore[valid-type]


class RoleTypes(StrEnum):
    domain_expert = "domain expert"
    information_architect = "information architect"
    cdf_solution_architect = "cdf solution architect"


class RuleModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True,
        str_strip_whitespace=True,
        arbitrary_types_allowed=True,
        strict=False,
        extra="allow",
        use_enum_values=True,
    )
    validators_to_skip: set[str] = Field(default_factory=set, exclude=True)

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)


class URL(BaseModel):
    url: HttpUrl


class CoreMetadata(RuleModel):
    """
    Metadata model for data model

    """

    role: RoleTypes | None = Field(description=("Role of the person who creates the data model"), default=None)

    prefix: Prefix | None = Field(
        description=("This is used as a short form for namespace"),
        default=None,
    )

    namespace: Namespace | None = Field(
        description="This is used as RDF namespace for generation of semantic data model representation",
        min_length=1,
        max_length=2048,
        default=None,
    )

    space: Space | None = Field(
        description=("CDF space to which data model is suppose to be stored under"),
        default=None,
    )

    externalId: ExternalId | None = Field(
        description=("Data model external id when resolving rules as CDF data model"),
        default=None,
        min_length=1,
        max_length=255,
    )

    version: str | None = Field(
        description=("Data model version"),
        min_length=1,
        max_length=43,
        default=None,
    )

    name: str | None = Field(
        description=("Human readable name of the data model"),
        min_length=1,
        max_length=255,
        default=None,
    )

    description: Description | None = Field(
        description=("Description/definition of the data model"),
        default=None,
    )

    created: datetime | None = Field(
        description=("Date of the data model creation"),
        default=None,
    )

    updated: datetime | None = Field(
        description=("Date of the data model update"),
        default=None,
    )

    creator: str | list[str] | None = Field(
        description=("Creator(s) of the data model"),
        default=None,
    )

    contributor: str | list[str] | None = Field(
        description=("Contributor(s) of the data model"),
        default=None,
    )

    rights: str | None = Field(
        description=("Usage rights of the data model"),
        default=None,
    )

    license: str | None = Field(
        description=("License of the data model"),
        default=None,
    )

    def to_pandas(self) -> pd.Series:
        """Converts Metadata to pandas Series."""
        return pd.Series(self.model_dump())

    def _repr_html_(self) -> str:
        """Returns HTML representation of Metadata."""
        return self.to_pandas().to_frame("value")._repr_html_()  # type: ignore[operator]


class CoreRules(RuleModel):
    """
    Rules is a core concept in `neat`. This represents fusion of data model
    definitions and (optionally) the transformation rules used to transform the data/graph
    from the source representation to the target representation defined by the data model.
    The rules are defined in a Excel sheet and then parsed into a `Rules` object. The
    `Rules` object is then used to generate data model and the`RDF` graph made of data
    model instances.

    Args:
        metadata: Data model metadata
        classes: Classes defined in the data model
        properties: Class properties defined in the data model with accompanying transformation rules
                    to transform data from source to target representation
        prefixes: Prefixes used in the data model. Defaults to PREFIXES
        instances: Instances defined in the data model. Defaults to None
        validators_to_skip: List of validators to skip. Defaults to []
    """

    metadata: CoreMetadata


#     @property
#     def raw_tables(self) -> list[str]:
#         return list(
#                 for rule in self.properties.values()
#                 if rule.is_raw_lookup

#     @field_validator("instances", mode="before")
#     def none_as_empty_list(cls, value):
#         if value is None:

#     @field_validator("classes", mode="before")
#     def dict_to_classes_obj(cls, value: dict | Classes) -> Classes:
#         if not isinstance(value, dict):

#     @field_validator("properties", mode="before")
#     def dict_to_properties_obj(cls, value: dict | Properties) -> Properties:
#         if not isinstance(value, dict):

#     @model_validator(mode="after")
#     @skip_model_validator("validators_to_skip")
#     def update_prefix_version_entities(self) -> Self:

#         # update expected_value_types
#         for id_ in self.properties.keys():
#             # only update version of expected value type which are part of this data model
#             if (
#                 and self.properties[id_].expected_value_type.prefix == "undefined"
#             ):

#             if self.properties[id_].expected_value_type.prefix == "undefined":

#         # update container
#         for id_ in self.properties.keys():
#             # only update version of expected value type which are part of this data model
#             if (
#                 and cast(ContainerEntity, self.properties[id_].container).prefix == "undefined"
#             ):

#         # update parent classes
#         for id_ in self.classes.keys():
#             if self.classes[id_].parent_class:
#                 for parent_class in cast(list[ParentClass], self.classes[id_].parent_class):
#                     if parent_class.prefix == "undefined":
#                     if not parent_class.version:


#     @model_validator(mode="after")
#     @skip_model_validator("validators_to_skip")
#     def update_container_description_and_name(self):
#         for id_ in self.properties.keys():
#             if (
#                 and self.properties[id_].container.external_id in self.classes
#                 and self.properties[id_].container.space == self.metadata.space
#             ):
#                 self.properties[id_].container.description = self.classes[
#                 ].description

#                 self.properties[id_].container.name = self.classes[
#                 ].class_name

#     @model_validator(mode="after")
#     @skip_model_validator("validators_to_skip")
#     def add_missing_classes(self):
#         for property_ in self.properties.values():
#             if property_.class_id not in self.classes:
#                 self.classes[property_.class_id] = Class(

#     def update_prefix(self, prefix: str):
#         if prefix == self.metadata.prefix:

#             # update entity ids for expected_value_types and containers
#             for id_ in self.properties.keys():
#                 if self.properties[id_].expected_value_type.prefix == old_prefix:

#                 if (
#                     and cast(ContainerEntity, self.properties[id_].container).prefix == old_prefix
#                 ):

#             # update parent classes
#             for id_ in self.classes.keys():
#                 if self.classes[id_].parent_class:
#                     for parent_class in cast(list[ParentClass], self.classes[id_].parent_class):
#                         if parent_class.prefix == old_prefix:

#             # update prefixes

#     def update_space(self, space: str):
#         "Convenience method for updating prefix more intuitive to CDF users"

#     def update_version(self, version: str):
#         if version == self.metadata.version:
#             for id_ in self.properties.keys():
#                 if (
#                     and self.properties[id_].expected_value_type.version == old_version
#                 ):

#             for id_ in self.classes.keys():
#                 if self.classes[id_].parent_class:
#                     for parent_class in cast(list[ParentClass], self.classes[id_].parent_class):
#                         if parent_class.prefix == self.metadata.prefix and parent_class.version == old_version:

#     @validator("prefixes")
#     @skip_field_validator("validators_to_skip")
#     def are_prefixes_compliant(cls, value, values):
#         if ill_formed_prefixes := [
#             prefix for prefix, _ in value.items() if re.search(more_than_one_none_alphanumerics_regex, prefix)
#         ]:
#             raise exceptions.MoreThanOneNonAlphanumericCharacter(
#             ).to_pydantic_custom_error()
#         if ill_formed_prefixes := [
#             prefix for prefix, _ in value.items() if not re.match(prefix_compliance_regex, prefix)
#         ]:
#             raise exceptions.PrefixesRegexViolation(
#                 ill_formed_prefixes, prefix_compliance_regex
#             ).to_pydantic_custom_error()

#     @validator("prefixes")
#     @skip_field_validator("validators_to_skip")
#     def are_namespaces_compliant(cls, value, values):
#         for _, namespace in value.items():

#         if ill_formed_namespaces:

#     @validator("prefixes")
#     @skip_field_validator("validators_to_skip")
#     def add_data_model_prefix_namespace(cls, value, values):
#         if "metadata" not in values:
#         if "prefix" not in values["metadata"].dict():
#             raise exceptions.FiledInMetadataSheetMissingOrFailedValidation(
#             ).to_pydantic_custom_error()
#         if "namespace" not in values["metadata"].dict():
#             raise exceptions.FiledInMetadataSheetMissingOrFailedValidation(
#             ).to_pydantic_custom_error()


#     @property
#     def space(self) -> str:
#         """Returns data model space."""

#     @property
#     def external_id(self) -> str:
#         """Returns data model external."""

#     @property
#     def name(self) -> str:
#         """Returns data model name."""

#     def _repr_html_(self) -> str:
#         """Pretty display of the TransformationRules object in a Notebook"""
#         for key in ["creator", "contributor"]:


# class Resource(RuleModel):
#     """
#     Base class for resources that constitute data model (i.e., classes, properties)

#     Args:
#         description: The description of the resource.
#         cdf_resource_type: The CDF resource type to which resource resolves to
#         deprecated: Whether the resource is deprecated or not.
#         deprecation_date: The date when the resource was deprecated.
#         replaced_by: The resource that replaced this resource.
#         source: Source of information for given resource
#         source_entity_name: The name of the source entity that is closest to the resource being described.
#         match_type: The match type of the resource being described and the source entity.
#         comment: Additional comment about mapping between the resource being described and the source entity.

#     """

#     # Solution model

#     # Solution CDF resource, it is not needed when working with FDM, this is only for
#     # Classic CDF data model

#     # Advance data modeling: Keeping track if Resource got deprecated or not
#     # Todo: Remove. Not used, only added as placeholder for future use.

#     # Advance data modeling: Relation to existing resources for purpose of mapping
#     source: HttpUrl | None = Field(
#         ),
#     source_entity_name: str | None = Field(
#     match_type: str | None = Field(

#     @model_validator(mode="before")
#     def replace_float_nan_with_default(cls, values: dict) -> dict:


# class ResourceDict(BaseModel, Generic[T_Resource]):

#     def __getitem__(self, item: str) -> T_Resource:

#     def __setitem__(self, key: str, value: T_Resource):

#     def __contains__(self, item: str) -> bool:

#     def __len__(self) -> int:

#     def __iter__(self) -> Iterator[str]:  # type: ignore[override]

#     def values(self) -> ValuesView[T_Resource]:

#     def keys(self) -> KeysView[str]:

#     def items(self) -> ItemsView[str, T_Resource]:

#     def to_pandas(self, drop_na_columns: bool = True, include: list[str] | None = None) -> pd.DataFrame:
#         """Converts ResourceDict to pandas DataFrame."""
#         if drop_na_columns:
#         if include is not None:

#     def groupby(self, by: str) -> dict[str, ResourceDict[T_Resource]]:
#         for key, resource in self.data.items():
#             if value not in groups:

#     def _repr_html_(self) -> str:
#         """Returns HTML representation of ResourceDict."""


# class Class(Resource):
#     """
#     Base class for all classes that are part of the data model.

#     Args:
#         class_id: The class ID of the class.
#         class_name: The name of the class.
#         parent_class: The parent class of the class.
#     """

#     # Solution model
#     # Used for hierarchical data modeling, and inheritance/extension of CDF data model
#     # Todo: Remove? Does not seem to be used anywhere.
#     #  This is for CDF solution architecht setting view.filter in data modeling.

#     @model_validator(mode="before")
#     def replace_nan_floats_with_default(cls, values: dict) -> dict:

#     @validator("class_id", always=True)
#     @skip_field_validator("validators_to_skip")
#     def is_class_id_compliant(cls, value, values):
#         if re.search(more_than_one_none_alphanumerics_regex, value):
#         if not re.match(class_id_compliance_regex, value):
#             raise exceptions.ClassSheetClassIDRegexViolation(
#                 value, class_id_compliance_regex
#             ).to_pydantic_custom_error()

#     @validator("class_name", always=True)
#     def set_class_name_if_none(cls, value, values):
#         if value is None:
#             if "class_id" not in values:
#             warnings.warn(

#     @field_validator("parent_class", mode="before")
#     @skip_field_validator("validators_to_skip")
#     def parent_class_to_list_of_entities(cls, value, values):
#         if isinstance(value, str) and value:
#             for v in value.replace(", ", ",").split(","):
#                 if ENTITY_ID_REGEX_COMPILED.match(v) or VERSIONED_ENTITY_REGEX_COMPILED.match(v):
#                     # if all fails defaults "neat" object which ends up being updated to proper
#                     # prefix and version upon completion of Rules validation


#     @field_validator("parent_class", mode="after")
#     @skip_field_validator("validators_to_skip")
#     def is_parent_class_id_compliant(cls, value, values):
#         if isinstance(value, list):
#             if illegal_ids := [v for v in value if re.search(more_than_one_none_alphanumerics_regex, v.suffix)]:
#                 raise exceptions.MoreThanOneNonAlphanumericCharacter(
#                 ).to_pydantic_custom_error()
#             if illegal_ids := [v for v in value if not re.match(class_id_compliance_regex, v.suffix)]:
#                 for v in illegal_ids:
#                 raise exceptions.ClassSheetParentClassIDRegexViolation(
#                     illegal_ids, class_id_compliance_regex
#                 ).to_pydantic_custom_error()


# class Classes(ResourceDict[Class]):
#     """This represents a collection of classes that are part of the data model."""

#     ...


# class Property(Resource):
#     """
#     A property is a characteristic of a class. It is a named attribute of a class that describes a range of values
#     or a relationship to another class.

#     Args:
#         class_id: Class ID to which property belongs
#         property_id: Property ID of the property
#         property_name: Property name. Defaults to property_id
#         expected_value_type: Expected value type of the property
#         min_count: Minimum count of the property values. Defaults to 0
#         max_count: Maximum count of the property values. Defaults to None
#         default: Default value of the property
#         property_type: Property type (DatatypeProperty/attribute or ObjectProperty/edge/relationship)
#         cdf_resource_type: CDF resource to under which property will be resolved to (e.g., Asset)
#         resource_type_property: To what property of CDF resource given property resolves to (e.g., Asset name)
#         source_type: In case if property resolves as CDF relationship, this argument indicates
#                      relationship source type (defaults to Asset)
#         target_type: In case if property resolves as CDF relationship, this argument
#                      indicates relationship target type (defaults to Asset)
#         label: CDF Label used for relationship. Defaults to property_id
#         relationship_external_id_rule: Rule to use when generating CDF relationship externalId
#         rule_type: Rule type for the transformation from source to target representation
#                    of knowledge graph. Defaults to None (no transformation)
#         rule: Actual rule for the transformation from source to target representation of
#               knowledge graph. Defaults to None (no transformation)
#         skip_rule: Flag indicating if rule should be skipped when performing
#                    knowledge graph transformations. Defaults to False

#     """

#     # Solution model

#     # OWL property

#     # Core CDF resources (Asset, Relationship, and Labels)
#     resource_type_property: list[str] | None = Field(
#         "example f cdf_resource_type is 'Asset', then this could"
#         "be 'name' or 'description'. Note you can specify "
#         "this property twice in CDF, once as 'name' and once as 'metadata",
#     # Specialization of cdf_resource_type to allow definition of both
#     # Asset and Relationship at the same time
#     cdf_resource_type: list[str] = Field(

#     # Transformation rule (domain to solution)
#     rule: str | AllReferences | SingleProperty | Hop | RawLookup | SPARQLQuery | Traversal | None = Field(

#     # Container specific things, only used for advance modeling or auto-filled by neat

#     @property
#     def is_raw_lookup(self) -> bool:

#     @model_validator(mode="before")
#     def replace_float_nan_with_default(cls, values: dict) -> dict:

#     @field_validator("container", mode="before")
#     def container_string_to_entity(cls, value):
#         if not value:

#         if ENTITY_ID_REGEX_COMPILED.match(value) or VERSIONED_ENTITY_REGEX_COMPILED.match(value):

#     @field_validator("expected_value_type", mode="before")
#     def expected_value_type_string_to_entity(cls, value):
#         # handle simple types
#         if value in XSD_VALUE_TYPE_MAPPINGS.keys():

#         # complex types correspond to relations to other classes
#         if ENTITY_ID_REGEX_COMPILED.match(value) or VERSIONED_ENTITY_REGEX_COMPILED.match(value):
#             return ValueType(
#         #     return ValueType(

#     @validator("class_id", always=True)
#     @skip_field_validator("validators_to_skip")
#     def is_class_id_compliant(cls, value, values):
#         if re.search(more_than_one_none_alphanumerics_regex, value):
#         if not re.match(class_id_compliance_regex, value):
#             raise exceptions.PropertiesSheetClassIDRegexViolation(
#                 value, class_id_compliance_regex
#             ).to_pydantic_custom_error()

#     @validator("property_id", always=True)
#     @skip_field_validator("validators_to_skip")
#     def is_property_id_compliant(cls, value, values):
#         if re.search(more_than_one_none_alphanumerics_regex, value):
#         if not re.match(property_id_compliance_regex, value):

#     @validator("expected_value_type", always=True)
#     @skip_field_validator("validators_to_skip")
#     def is_expected_value_type_compliant(cls, value, values):
#         if re.search(more_than_one_none_alphanumerics_regex, value.suffix):
#             raise exceptions.MoreThanOneNonAlphanumericCharacter(
#                 "expected_value_type", value
#             ).to_pydantic_custom_error()
#         if not re.match(class_id_compliance_regex, value.suffix):

#     @validator("rule_type", pre=True)
#     def to_lowercase(cls, value):

#     @validator("skip_rule", pre=True)
#     def from_string(cls, value):
#         if isinstance(value, str):

#     @validator("rule")
#     @skip_field_validator("validators_to_skip")
#     def is_valid_rule(cls, value, values):
#         if rule_type := values.get("rule_type"):
#             if not value:
#                 raise exceptions.RuleTypeProvidedButRuleMissing(
#                 ).to_pydantic_custom_error()

#     @validator("resource_type_property", pre=True)
#     def split_str(cls, v):
#         if v:

#     @field_validator("cdf_resource_type", mode="before")
#     def to_list_if_comma(cls, value, info):
#         if isinstance(value, str):
#             if value:
#             if cls.model_fields[info.field_name].default is None:

#     # Setters
#     # TODO: configure setters to only run if field_validators are successful, otherwise do not run them!
#     @property
#     def is_mandatory(self) -> bool:

#     @model_validator(mode="after")
#     def set_property_type(self):
#         if self.expected_value_type.type_ == EntityTypes.data_value_type:

#     @model_validator(mode="after")
#     def set_container_if_missing(self):
#         if not self.container and (
#         ):

#     @model_validator(mode="after")
#     def set_container_property_if_missing(self):
#         if not self.container_property and (
#         ):

#     @model_validator(mode="after")
#     def set_property_name_if_none(self):
#         if self.property_name is None:
#             warnings.warn(

#     @model_validator(mode="after")
#     @skip_model_validator("validators_to_skip")
#     def set_relationship_label(self):
#         if self.label is None:
#             warnings.warn(

#     @model_validator(mode="after")
#     @skip_model_validator("validators_to_skip")
#     def set_skip_rule(self):
#         if self.rule_type is None:
#             warnings.warn(

#     @model_validator(mode="after")
#     def set_default_as_list(self):
#         if (
#             and self.default
#             and self.max_count
#             and self.max_count != 1
#             and not isinstance(self.default, list)
#         ):
#             warnings.warn(
#             if isinstance(self.default, str):
#                 if self.default:

#     @model_validator(mode="after")
#     @skip_model_validator("validators_to_skip")
#     def is_default_value_type_proper(self):
#         if self.property_type == "DatatypeProperty" and self.default:

#             if type(default_value) != self.expected_value_type.python:
#                     if isinstance(self.default, list):
#                         for value in self.default:

#                     exceptions.DefaultValueTypeNotProper(
#                         self.property_id,
#                         self.expected_value_type.python,


# class Properties(ResourceDict[Property]):
#     """This represents a collection of properties that are part of the data model."""

#     ...


# class Prefixes(RuleModel):
#     """
#     Class deals with prefixes used in the data model and data model instances

#     Args:
#         prefixes: Dict of prefixes
#     """


# class Instance(RuleModel):
#     """
#     Class deals with RDF triple that defines data model instances of data, represented
#     as a single row in the `Instances` sheet of the Excel file.

#     Args:
#         instance: URI of the instance
#         property_: URI of the property
#         value: value of the property
#         namespace: namespace of the instance
#         prefixes: prefixes of the instance

#     !!! note "Warning"
#         Use of the `Instances` sheet is not recommended, instead if you need additional
#         triples in your graph use Graph Capturing Sheet instead!

#         See
#         [`extract_graph_from_sheet`](../graph/extractors.md#cognite.neat.graph.extractors.extract_graph_from_sheet)
#         for more details.
#     """


#     @staticmethod
#     def get_value(value, prefixes) -> URIRef | Literal:

#     @model_validator(mode="before")
#     def convert_values(cls, values: dict):
#         # we expect to read Excel sheet which contains naming convention of column
#         # 'Instance', 'Property', 'Value', if that's not the case we should raise error
#         if not {"Instance", "Property", "Value"}.issubset(set(values.keys())):


#         values["Instance"] = (

#         values["Property"] = (

#         if isinstance(values["Value"], str):
#             if not isinstance(values["Value"], URIRef):
#                     XSD.integer
#                     if cls.isint(values["Value"])
#                     else XSD.float
#                     if cls.isfloat(values["Value"])
#                     else XSD.string


#     @staticmethod
#     def isfloat(x):

#     @staticmethod
#     def isint(x):



from datetime import datetime

from rdflib import Namespace

from .base import Prefix, RoleTypes, RuleModel
from .domain_rules import DomainMetadata


class InformationArchitectMetadata(DomainMetadata):
    role: RoleTypes = RoleTypes.information_architect
    prefix: Prefix
    namespace: Namespace
    version: str
    contributor: str | list[str]
    created: datetime
    updated: datetime

    @classmethod
    def from_domain_expert_metadata(
        cls,
        metadata: DomainMetadata,
        prefix: str | None = None,
        namespace: Namespace | None = None,
        version: str | None = None,
        contributor: str | list[str] | None = None,
        created: datetime | None = None,
        updated: datetime | None = None,
    ):
        metadata_as_dict = metadata.model_dump()
        metadata_as_dict["prefix"] = prefix or metadata.prefix or "neat"
        metadata_as_dict["namespace"] = namespace or metadata.namespace or Namespace("http://purl.org/cognite/neat#")
        metadata_as_dict["version"] = version or metadata.version or "0.1.0"
        metadata_as_dict["contributor"] = contributor or metadata.contributor or "Cognite"
        metadata_as_dict["created"] = created or metadata.created or datetime.utcnow()
        metadata_as_dict["updated"] = updated or metadata.updated or datetime.utcnow()
        return cls(**metadata_as_dict)


class InformationArchitectRules(RuleModel):
    metadata: InformationArchitectMetadata



"""Exports rules to CDF Data Model Storage (DMS) through cognite-sdk.
"""

import dataclasses
import logging
import sys
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Literal, cast

import yaml

from cognite.neat.rules.models.value_types import ValueTypeMapping

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import (
    ContainerApply,
    ContainerApplyList,
    ContainerId,
    ContainerProperty,
    DataModelApply,
    DataModelId,
    DirectRelation,
    DirectRelationReference,
    MappedPropertyApply,
    PropertyType,
    SpaceApply,
    ViewApply,
    ViewId,
)
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType
from cognite.client.data_classes.data_modeling.views import (
    ConnectionDefinitionApply,
    SingleHopConnectionDefinitionApply,
)
from pydantic import BaseModel, ConfigDict, field_validator

from cognite.neat.rules import exceptions
from cognite.neat.rules.exporter._base import BaseExporter
from cognite.neat.rules.exporter._validation import are_entity_names_dms_compliant
from cognite.neat.rules.models._base import ContainerEntity, EntityTypes, ParentClass
from cognite.neat.rules.models.rules import Class, Property, Rules
from cognite.neat.utils.utils import generate_exception_report

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup


@dataclass
class EmptyPropertyType(PropertyType):
    _type = "No property"


@dataclass
class DMSSchema:
    data_model: DataModelApply
    containers: ContainerApplyList


class DMSExporter(BaseExporter[DMSSchema]):
    """Class for exporting transformation rules object to CDF Data Model Storage (DMS).

    Args:
        rules: Transformation rules object.
        data_model_id: The id of the data model to be created.
        container_policy: How to create/reuse existing containers.
        existing_model: In the case of updating an existing model, this is the existing model.
        report: Report. This is used when the exporter object is created from RawRules

    !!! note "Container policy"
        Here is more information about the different container policies:
        - `create`: assumes no containers exist in CDF, will attempt to create them
        - `reuse`: assumes containers exists in CDF, will attempt to only re-use them
        - `extend`: will re-use existing, extend and/or create only missing containers
        - `optimize`: create containers of size's prescribed by NEAT's optimization algorithm
    """

    def __init__(
        self,
        rules: Rules,
        data_model_id: dm.DataModelId | None = None,
        container_policy: Literal["create", "reuse", "extend", "optimize"] = "create",
        existing_model: dm.DataModel[dm.View] | None = None,
        report: str | None = None,
    ):
        super().__init__(rules, report)
        self.data_model_id = data_model_id
        self.container_policy = container_policy
        if container_policy == "extend" and existing_model is None:
            raise ValueError("Container policy is extend-existing, but no existing model is provided")
        if container_policy != "create":
            raise NotImplementedError("Only one-to-one-view container policy is currently supported")
        self.existing_model = existing_model

    def _export_to_file(self, filepath: Path) -> None:
        if filepath.suffix not in {".yaml", ".yml"}:
            warnings.warn("File extension is not .yaml, adding it to the file name", stacklevel=2)
            filepath = filepath.with_suffix(".yaml")
        schema = self.export()
        filepath.write_text(
            yaml.safe_dump(
                {
                    "data_models": [schema.data_model.dump(camel_case=True)],
                    "containers": schema.containers.dump(camel_case=True),
                }
            )
        )

    def export(self) -> DMSSchema:
        model = DataModel.from_rules(self.rules, self.data_model_id)
        return DMSSchema(
            data_model=DataModelApply(
                space=model.space,
                external_id=model.external_id,
                version=model.version,
                description=model.description,
                name=model.name,
                views=list(model.views.values()),
            ),
            containers=ContainerApplyList(model.containers.values()),
        )


class DataModel(BaseModel):
    """
    Data model pydantic class used to create space, containers, views and data model in CDF.

    This can be used to create a data model in CDF from transformation rules.

    Args:
        space: Name of the space to place the resulting data model.
        external_id: External id of the data model.
        version: Version of the data model.
        description: Description of the data model.
        name: Name of the data model.
        containers: Containers connected to the data model.
        views: Views connected to the data model.

    """

    space: str
    external_id: str
    version: str
    description: str | None = None
    name: str | None = None
    containers: dict[str, ContainerApply]
    views: dict[str, ViewApply]

    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, arbitrary_types_allowed=True, strict=False, extra="allow"
    )

    @field_validator("views", mode="after")
    def remove_empty_views(cls, value):
        """This validator removes views that do not have any properties or implements."""
        for view_id, view in list(value.items()):
            if not view.properties and not view.implements:
                del value[view_id]

        return value

    @property
    def spaces(self):
        return set(
            [container.space for container in self.containers.values()]
            + [self.space]
            + [view.space for view in self.views.values()]
        )

    @classmethod
    def from_rules(
        cls, rules: Rules, data_model_id: dm.DataModelId | None = None, set_expected_source: bool = True
    ) -> Self:
        """Generates a DataModel class instance from a Rules instance.

        Args:
            rules: instance of Rules.
            data_model_id: The id of the data model to be created.
            set_expected_source: Whether to set the expected source on views with direct relation properties.

        Returns:
            Instance of DataModel.
        """
        if data_model_id and data_model_id.space:
            # update space in rules to match space
            rules.update_space(data_model_id.space)

        if data_model_id and data_model_id.version:
            # update version in rules to match version
            rules.update_version(data_model_id.version)

        if data_model_id and data_model_id.external_id:
            # update data model name to match external_id
            rules.metadata.suffix = data_model_id.external_id

        names_compliant, name_warnings = are_entity_names_dms_compliant(rules, return_report=True)
        if not names_compliant:
            raise exceptions.EntitiesContainNonDMSCompliantCharacters(report=generate_exception_report(name_warnings))

        if rules.metadata.external_id is None:
            raise exceptions.DataModelIdMissing(prefix=rules.metadata.space)

        return cls(
            space=rules.metadata.space,
            external_id=rules.metadata.external_id,
            version=rules.metadata.version,
            description=rules.metadata.description,
            name=rules.metadata.name,
            containers=cls.containers_from_rules(rules),
            views=cls.views_from_rules(rules),
        )

    @staticmethod
    def containers_from_rules(rules: Rules) -> dict[str, ContainerApply]:
        """Create a dictionary of ContainerApply instances from a Rules instance.

        Args:
            rules: instance of Rules.`

        Returns:
            Dictionary of ContainerApply instances.
        """

        containers: dict[str, ContainerApply] = {}
        errors: list = []

        for row, property_ in rules.properties.items():
            if property_.container:
                container_property_id: str = cast(str, property_.container_property)
                container_id = property_.container.id

                api_container = DataModel.as_api_container(property_.container)
                api_container_property = DataModel.as_api_container_property(property_)

                # scenario: adding new container to the data model for the first time
                if not isinstance(api_container_property.type, EmptyPropertyType) and container_id not in containers:
                    containers[container_id] = api_container
                    containers[container_id].properties[container_property_id] = api_container_property

                # scenario: adding new property to an existing container
                elif (
                    not isinstance(api_container_property.type, EmptyPropertyType)
                    and container_property_id not in containers[container_id].properties
                ):
                    containers[container_id].properties[container_property_id] = api_container_property

                # scenario: property is redefined checking for potential sub-scenarios
                elif (
                    not isinstance(api_container_property.type, EmptyPropertyType)
                    and container_property_id in containers[container_id].properties
                ):
                    existing_property = containers[container_id].properties[container_property_id]

                    # scenario: property is redefined with a different value type -> raise error
                    if not isinstance(existing_property.type, type(api_container_property.type)):
                        errors.append(
                            exceptions.ContainerPropertyValueTypeRedefinition(
                                container_id=container_id,
                                property_id=container_property_id,
                                current_value_type=str(existing_property.type),
                                redefined_value_type=str(api_container_property.type),
                                loc=f"[Properties/Type/{row}]",
                            )
                        )

                    # scenario: default value is redefined -> set default value to None
                    if (
                        not isinstance(existing_property.type, DirectRelation)
                        and not isinstance(api_container_property.type, DirectRelation)
                        and existing_property.default_value != api_container_property.default_value
                    ):
                        containers[container_id].properties[container_property_id] = dataclasses.replace(
                            existing_property, default_value=None
                        )

                    # scenario: property hold multiple values -> set is_list to True
                    if (
                        not isinstance(existing_property.type, DirectRelation)
                        and not isinstance(api_container_property.type, DirectRelation)
                        and cast(ListablePropertyType, existing_property.type).is_list
                        != cast(ListablePropertyType, api_container_property.type).is_list
                    ):
                        cast(
                            ListablePropertyType, containers[container_id].properties[container_property_id].type
                        ).is_list = True

        if errors:
            raise ExceptionGroup("Properties value types have been redefined! This is prohibited! Aborting!", errors)

        return containers

    @staticmethod
    def as_api_container_property(property_: Property) -> ContainerProperty:
        is_one_to_many = property_.max_count != 1

        # Literal, i.e. Node attribute
        if property_.property_type is EntityTypes.data_property:
            property_type = cast(
                type[ListablePropertyType],
                cast(ValueTypeMapping, property_.expected_value_type.mapping).dms,
            )
            return ContainerProperty(
                type=property_type(is_list=is_one_to_many),
                nullable=property_.min_count == 0,
                default_value=property_.default,
                name=property_.property_name,
                description=property_.description,
            )

        # URIRef, i.e. edge 1-1 , aka direct relation between Nodes
        elif property_.property_type is EntityTypes.object_property and not is_one_to_many:
            return ContainerProperty(
                type=DirectRelation(),
                nullable=True,
                name=property_.property_name,
                description=property_.description,
            )

        # return type=None if property cannot be created
        else:
            return ContainerProperty(type=EmptyPropertyType())

    @staticmethod
    def as_api_container(container: ContainerEntity) -> ContainerApply:
        return ContainerApply(
            space=container.space,
            external_id=container.external_id,
            description=container.description,
            name=container.name,
            # properties are added later
            properties={},
        )

    @staticmethod
    def views_from_rules(rules: Rules) -> dict[str, ViewApply]:
        """Generates a dictionary of ViewApply instances from a Rules instance.

        Args:
            rule: Instance of Rules.
            space: Name of the space to place the views.

        Returns:
            Dictionary of ViewApply instances.
        """
        views: dict[str, ViewApply] = {
            f"{rules.metadata.space}:{class_.class_id}": DataModel.as_api_view(
                class_, rules.metadata.space, rules.metadata.version
            )
            for class_ in rules.classes.values()
        }
        errors: list = []

        # Create views from property-class definitions
        for row, property_ in rules.properties.items():
            view_property = DataModel.as_api_view_property(property_, rules.metadata.space)
            id_ = f"{rules.metadata.space}:{property_.class_id}"

            # scenario: view exist but property does not so it is added
            if view_property and (property_.property_id not in cast(dict, views[id_].properties)):
                cast(dict, views[id_].properties)[property_.property_id] = view_property

            # scenario: view exist, property exists but it is differently defined -> raise error
            # type: ignore
            elif (
                view_property
                and property_.property_id in cast(dict, views[id_].properties)
                and view_property is not cast(dict, views[id_].properties)[property_.property_id]
            ):
                errors.append(
                    exceptions.ViewPropertyRedefinition(
                        view_id=id_,
                        property_id=cast(str, property_.property_id),
                        loc=f"[Properties/Property/{row}]",
                    )
                )

        if errors:
            raise ExceptionGroup("View properties have been redefined! This is prohibited! Aborting!", errors)

        return views

    @staticmethod
    def as_api_view(class_: Class, space: str, version: str) -> ViewApply:
        return ViewApply(
            space=space,
            external_id=class_.class_id,
            version=version,
            name=class_.class_name,
            description=class_.description,
            properties={},
            implements=[parent_class.view_id for parent_class in cast(list[ParentClass], class_.parent_class)]
            if class_.parent_class
            else None,
        )

    @staticmethod
    def as_api_view_property(property_: Property, space: str) -> MappedPropertyApply | ConnectionDefinitionApply | None:
        property_.container = cast(ContainerEntity, property_.container)
        property_.container_property = cast(str, property_.container_property)
        if property_.property_type is EntityTypes.data_property:
            return MappedPropertyApply(
                container=ContainerId(space=property_.container.space, external_id=property_.container.external_id),
                container_property_identifier=property_.container_property,
                name=property_.property_name,
                description=property_.description,
            )

        # edge 1-1 == directRelation
        elif property_.property_type is EntityTypes.object_property and property_.max_count == 1:
            return MappedPropertyApply(
                container=ContainerId(space=property_.container.space, external_id=property_.container.external_id),
                container_property_identifier=property_.container_property,
                name=property_.property_name,
                description=property_.description,
                source=ViewId(
                    space=property_.expected_value_type.space,
                    external_id=property_.expected_value_type.external_id,
                    version=property_.expected_value_type.version,
                ),
            )

        # edge 1-many == EDGE, this does not have container! type is here source ViewId ?!
        elif property_.property_type is EntityTypes.object_property and property_.max_count != 1:
            return SingleHopConnectionDefinitionApply(
                type=DirectRelationReference(space=space, external_id=f"{property_.class_id}.{property_.property_id}"),
                # Here we create ViewID to the container that the edge is pointing to.
                source=ViewId(
                    space=property_.expected_value_type.space,
                    external_id=property_.expected_value_type.external_id,
                    version=property_.expected_value_type.version,
                ),
                direction="outwards",
                name=property_.property_name,
                description=property_.description,
            )
        else:
            return None

    def to_cdf(self, client: CogniteClient):
        """Write the the data model to CDF.

        Args:
            client: Connected Cognite client.

        """
        existing_data_model = self.find_existing_data_model(client)
        existing_containers = self.find_existing_containers(client)
        existing_views = self.find_existing_views(client)

        if existing_data_model or existing_containers or existing_views:
            raise exceptions.DataModelOrItsComponentsAlreadyExist(
                existing_data_model, existing_containers, existing_views
            )

        self.create_space(client)
        self.create_containers(client)
        self.create_views(client)
        self.create_data_model(client)

    def find_existing_data_model(self, client: CogniteClient) -> DataModelId | None:
        """Checks if the data model exists in CDF.

        Args:
            client: Cognite client.

        Returns:
             True if the data model exists, False otherwise.
        """
        if model := client.data_modeling.data_models.retrieve((self.space, self.external_id, self.version)):
            cdf_data_model = model.latest_version()
            logging.warning(exceptions.DataModelAlreadyExist(self.external_id, self.version, self.space).message)
            warnings.warn(
                exceptions.DataModelAlreadyExist(self.external_id, self.version, self.space).message,
                category=exceptions.DataModelAlreadyExist,
                stacklevel=2,
            )

            return cdf_data_model.as_id()
        return None

    def find_existing_containers(self, client: CogniteClient) -> set:
        """Checks if the containers exist in CDF.

        Args:
            client: Cognite client.

        Returns:
            True if the containers exist, False otherwise.
        """

        existing_containers = set()

        for id_, container in self.containers.items():
            if client.data_modeling.containers.retrieve(container.as_id()):
                existing_containers.add(id_)

        return existing_containers

    def find_existing_views(self, client: CogniteClient) -> set:
        """Checks if the views exist in CDF.

        Args:
            client: Cognite client.

        Returns:
            True if the views exist, False otherwise.
        """

        existing_views = set()

        for id_, view in self.views.items():
            if client.data_modeling.views.retrieve(view.as_id()):
                existing_views.add(id_)

        return existing_views

    def create_space(self, client: CogniteClient):
        logging.info(f"Creating space {self.space}")

        for space in self.spaces:
            _ = client.data_modeling.spaces.apply(SpaceApply(space=space))

    def create_containers(self, client: CogniteClient):
        for container_id in self.containers:
            logging.info(f"Creating container {container_id} in space {self.space}")
            _ = client.data_modeling.containers.apply(self.containers[container_id])

    def create_views(self, client: CogniteClient):
        for view_id in self.views:
            logging.info(f"Creating view {view_id} version {self.views[view_id].version} in space {self.space}")
            _ = client.data_modeling.views.apply(self.views[view_id])

    def create_data_model(self, client: CogniteClient):
        logging.info(f"Creating data model {self.external_id} version {self.version} in space {self.space}")
        _ = client.data_modeling.data_models.apply(
            DataModelApply(
                name=self.name,
                description=self.description,
                space=self.space,
                external_id=self.external_id,
                version=self.version,
                views=list(self.views.values()),
            )
        )

    def remove_data_model(self, client: CogniteClient):
        """Helper function to remove a data model, and all underlying views and containers from CDF.

        Args:
            client: Cognite client.
        """

        if client.data_modeling.data_models.retrieve((self.space, self.external_id, self.version)):
            logging.info(f"Removing data model {self.external_id} version {self.version} from space {self.space}")
            _ = client.data_modeling.data_models.delete((self.space, self.external_id, self.version))

        if views := client.data_modeling.views.retrieve(
            [view.as_id() for view in self.views.values()], all_versions=False
        ):
            for view in views:
                logging.info(f"Removing view {view.external_id} version {view.version} from space {self.space}")
                _ = client.data_modeling.views.delete((view.space, view.external_id, view.version))

        if containers := client.data_modeling.containers.retrieve(
            [container.as_id() for container in self.containers.values()]
        ):
            for container in containers:
                logging.info(f"Removing container {container.external_id} from space {self.space}")
                _ = client.data_modeling.containers.delete((container.space, container.external_id))



from datetime import datetime

from cognite.neat.rules.models._rules.information_rules import InformationArchitectMetadata

from .base import CoreRules, ExternalId, RoleTypes, Space
from .domain_rules import DomainMetadata


class AssetSolutionArchitectMetadata(InformationArchitectMetadata):
    ...


class DmsSolutionArchitectMetadata(DomainMetadata):
    role: RoleTypes = RoleTypes.cdf_solution_architect
    space: Space
    externalId: ExternalId
    version: str
    contributor: str | list[str]
    created: datetime
    updated: datetime

    @classmethod
    def from_information_architect_metadata(
        cls, metadata: InformationArchitectMetadata, space: str | None = None, externalId: str | None = None
    ):
        metadata_as_dict = metadata.model_dump()
        metadata_as_dict["space"] = space or metadata.space or "neat-playground"
        metadata_as_dict["externalId"] = externalId or metadata.externalId or "neat_model"
        return cls(**metadata_as_dict)

    @classmethod
    def from_domain_expert_metadata(
        cls,
        metadata: DomainMetadata,
        space: str | None = None,
        externalId: str | None = None,
        version: str | None = None,
        contributor: str | list[str] | None = None,
        created: datetime | None = None,
        updated: datetime | None = None,
    ):
        metadata_as_dict = metadata.model_dump()
        metadata_as_dict["space"] = space or metadata.space or "neat-playground"
        metadata_as_dict["externalId"] = externalId or metadata.externalId or "neat_model"
        metadata_as_dict["version"] = version or metadata.version or "0.1.0"
        metadata_as_dict["contributor"] = contributor or metadata.contributor or "Cognite"
        metadata_as_dict["created"] = created or metadata.created or datetime.utcnow()
        metadata_as_dict["updated"] = updated or metadata.updated or datetime.utcnow()

        return cls(**metadata_as_dict)


class AssetRules(CoreRules):
    metadata: AssetSolutionArchitectMetadata


class DMSRules(CoreRules):
    metadata: DmsSolutionArchitectMetadata
